// definitions for segdwt
#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <sstream>
#include <cmath>
#include "segdwt.h"

// ostream overloads
ostream& operator<<(ostream& stream, vecType& vec) {
	vecIter iter = vec.begin();
	stream << "Vector output (size = " << vec.size() << ")" << std::fixed << std::setprecision(3) << std::endl;
	while(iter != vec.end()) {
		stream << *iter++ << "\t";
	}
	stream << std::endl;	
	return stream; 
}


// SegDWT class definitions
// constructor
SegDWT::SegDWT(unsigned segmentSize, unsigned depth, bool debg) : segSize_(segmentSize), wDepth_(depth), debug_(debg) {
	twoToJ_ = (int) pow((float) 2, (float) wDepth_);
}
// load filter specified by filename
// exception-based model of error-treating
void SegDWT::loadFilter(const char *filename) {
	std::fstream file;
	file.open(filename);
	if(!file) {
		std::cout << "Can't open filter file!" << std::endl;
		throw ExcFilterFile();
	}
	
	// process all lines of text
	unsigned filterType = 0;
	std::string line;
	
	while(!std::getline(file, line).eof() && filterType < 4) {
		type value=0.0;
		std::istringstream strCheck(line);
		
		// check text line
		if(strCheck >> value) {
			switch(filterType) {
				case 0:
					LoDFilter_.push_back(value);
					break;
				case 1:
					HiDFilter_.push_back(value);
					break;
				case 2:
					LoRFilter_.push_back(value);
					break;
				case 3:
					HiRFilter_.push_back(value);
					break;
					
			}
		} else {
			filterType++;
		}
	}
	
	// check last vector
	if(HiRFilter_.size() == 0 || LoDFilter_.size() == 0 || LoRFilter_.size() == 0 || HiDFilter_.size() == 0) {
		std::cout << "Filter file in bad format or incomplete!" << std::endl;
		throw ExcFilterFile();
	}
	
	std::cout << "Filter file \"" << filename << "\" loaded. Coefs count: "
				<< "LoD=" << LoDFilter_.size() << ", "
				<< "HiD=" << HiDFilter_.size() << ", "
				<< "LoR=" << LoRFilter_.size() << ", "
				<< "HiR=" << HiRFilter_.size() << std::endl;	  
}

// load signal file specified by filename
// exception-based model of error-treating
void SegDWT::loadSignal(const char *filename) {
	std::fstream file;
	file.open(filename);
	if(!file) {
		std::cout << "Can't open signal file!" << std::endl;
		throw ExcSignalFile();
	}
	
	// process all lines of text
	std::string line;
	
	while(!std::getline(file, line).eof()) {
		type value=0.0;
		std::istringstream strCheck(line);
		
		// check text line
		if(strCheck >> value) {
			signal_.push_back(value);
		} else {
			break;
		}
	}
	
	if(signal_.size() == 0) {
		std::cout << "Signal file in wrong format, no samples were detected!" << std::endl;
		throw ExcSignalFile();
	}
		
	std::cout << "Signal file \"" << filename << "\" loaded. Size of signal is " << signal_.size() << " samples." << std::endl;	  
	
	segCount_ = (int) ceil((float) signal_.size() / (float) segSize_);
	lastSegSize_ = (signal_.size() % segSize_ == 0) ? segSize_ : signal_.size() % segSize_;
	
	std::cout << "Estimated segment count: " << segCount_ << ", last segment size=" << lastSegSize_ << std::endl;
}
// save signal specified by filename
// format:
// -------------------------------
// SEGMENTS(=s)  WDEPTH(=d)
// S1Cnt  D11Cnt ... D1dCnt A1dCnt 
// S2Cnt  D21Cnt ... D2dCnt A2dCnt
// ...    ...    ... ...    ...
// SsCnt  Ds1Cnt ... DsdCnt AsdCnt
// 
// D11 data separated by <enter>
// 
// D12 data separated by <enter>
// ...
// 
// D1d data separated by <enter>
// 
// A1d data separated by <enter>
// 
// D21 data ...
// -------------------------------
void SegDWT::saveResult(const char *filename) {
	std::ofstream file;

	file.open(filename, std::ios::out);

	file << segCount_ << "\t" << wDepth_ << std::endl;

	std::list<Segment>::iterator segmentIt = seg_.begin();
	while(segmentIt != seg_.end()) {
		std::list<Segment::WaveletCoefs>::iterator coefsIt = segmentIt->coefs.begin();
		// print segment size
		file << ((segmentIt->isLast)?lastSegSize_:segSize_) << "\t";
		// print coef sizes
		unsigned depth=0;
		while(coefsIt != segmentIt->coefs.end()) {
			file << coefsIt->d.size() << "\t";
			if(depth == wDepth_-1) {
				file << coefsIt->a.size();
				break;
			}	
			depth++; coefsIt++;
		}
		file << std::endl;
		segmentIt++;
	}
	
	file << std::setprecision(15) << std::fixed;
	
	// save all coefs
	segmentIt = seg_.begin();
	while(segmentIt != seg_.end()) {
		std::list<Segment::WaveletCoefs>::iterator coefsIt = segmentIt->coefs.begin();
		// extra enter

		// print segment values
		unsigned depth=0;
		while(coefsIt != segmentIt->coefs.end()) {
			// print D coefs		
			file << std::endl;
			for(unsigned j=0; j<coefsIt->d.size(); ++j) {
				file << coefsIt->d[j] << std::endl;
			}
			// print last A coefs
			if(depth == wDepth_-1) {
				file << std::endl;
				for(unsigned j=0; j<coefsIt->d.size(); ++j) {
					file << coefsIt->a[j] << std::endl;
				}	
				break;
			}	
			depth++; coefsIt++;
		}
		segmentIt++;
	}
	file.close();
	std::cout << "Results saved to file \"" << filename << "\"." << std::endl;
}
// proceed with transform
void SegDWT::performTransform() {
	n_ = 0;
	extLen_ = getLeftExtSize(0);
	
	// normal segment processing
	while(n_ < segCount_ - 1) {
		if(DEBUG) {
			std::cout << std::endl << "Segment: " << n_ << " ---------------------" << std::endl;
			std::cout << "extL(" << n_+1 <<  ") = " << getLeftExtSize(n_) << std::endl; 
			std::cout << "extR(" << n_+1 <<  ") = " << getRightExtSize(n_) << std::endl;
		}
		
		// assign segment to list
		seg_.push_back(Segment(wDepth_, false));
		
		// init temporary segment data
		vecType seg;
		vecIter segBeg = signal_.begin() + (segSize_*n_);
		
		// copy temporary segment from signal and extend it
		seg.assign(segBeg, segBeg + segSize_);
		extendLeft(n_, seg);
		extendRight(n_, 0, seg);
		
		if(DEBUG) {
			std::cout << "SegData: " << seg << std::endl;
		}
		
		mlwd(seg_.back(), seg);
				
		n_++;
	}
	
	// perform computation of last segment right extension
	int zerosR = segSize_ - ((getLeftExtSize(n_) + lastSegSize_) - ((n_==0) ? extLen_ : getLeftExtSize(n_)));
	// and secondary extension by twoToJ
	zerosR += extLen_;
	
	// last segment processing	
	if(DEBUG) {
		std::cout << std::endl << "Segment: " << n_ << " (LAST) ---------------------" << std::endl;
		std::cout << "extL(" << n_+1 <<  ") = " << getLeftExtSize(n_) << std::endl; 
		std::cout << "extR(" << n_+1 <<  ") = " << zerosR << std::endl;
	}
	
	// add segment to list
	seg_.push_back(Segment(wDepth_, true));
	// create segment temp vector
	vecType seg;
	vecIter segBeg = signal_.begin() + segSize_*n_;
	
	// assign data & extend
	seg.assign(segBeg, signal_.end());
	extendLeft(n_, seg);
	extendRight(n_, zerosR, seg);
	
	if(DEBUG) {
		std::cout << "SegData: " << seg << std::endl;
	}
	
	// perform mldw
	mlwd(seg_.back(), seg);
}


//---- private methods
// get ext left size
int SegDWT::getLeftExtSize(int segment) {
	// any other
	if(segment > 0)
		return extLen_ - getRightExtSize(segment - 1);
	// first
	else
		return (twoToJ_-1) * (LoDFilter_.size()-1);
}
// get ext right size
int SegDWT::getRightExtSize(int segment) {
	// just to match original equation
	segment++;
	// compute right extension
	return twoToJ_ * (int) ceil((float) (segment * segSize_) / (float) twoToJ_) -  segment * segSize_;
} 
// !only for proper segments
// model extension from left
// params: segment - segment number, vector - a vector to be extended (&ref)
void SegDWT::extendLeft(int segment, vecType& vector) {
	// segment start computation
	int segStart = segment * segSize_ - getLeftExtSize(segment) - 1;
	// insert enough elements to vector beginning with 0.0 value
	vector.insert(vector.begin(), getLeftExtSize(segment), 0.0);
	
	for(int i=0; i<getLeftExtSize(segment); ++i) {
		segStart++;
		// copy of previous segment sample
		if(segStart >= 0)
			vector[i] = signal_[segStart];
	}
}
// model extension from left
// params: segment - segment number, specR - explicitly specified right extension
// (if you leave with 0, will recompute from getRightExtSize()), vector - a vector to be extended (&ref)
void SegDWT::extendRight(int segment, int specR, vecType& vector) {
	vecIter segEnd;
	
	// branch by last segment
	if(segment < (int) segCount_ - 1) {
		segEnd = signal_.begin() + segment * segSize_ + (segSize_ - 1);
	} else {
		segEnd = signal_.end() - 1;
	}
	
	// treating special case of last segment - data fetched by parameter
	if(specR == 0) {
		specR = getRightExtSize(segment);
	}
	
	for(int i=0; i<specR; ++i) {
		segEnd++;
		if(segEnd == signal_.end()) {
			// fill the rest with zeros and break
			if(specR-i > 1) {
				vector.push_back(0);
				if(specR-i > 2) {
					vector.insert(vector.end()-1, specR-i-1, 0.0);
				}
			}
			break;
		} else {
			// copy of next segment sample
			vector.push_back(*segEnd);
		}
	}	
}

// mldw algorithm
// params: &ref to Segment, &ref to source &vecType data
void SegDWT::mlwd(SegDWT::Segment &seg, vecType &src) {
	unsigned d=0;
	std::list<Segment::WaveletCoefs>::iterator depth = seg.coefs.begin();
	vecType input = src;
	
	// for every depth
	while(d < wDepth_ && depth != seg.coefs.end()) {
		// convolve
		conv(input, LoDFilter_, depth->a);
		conv(input, HiDFilter_, depth->d);
		// dyadic downgrade & cutoff by filterSize-1
		changeSignal(depth->a, dyaddown, LoDFilter_.size()-1);
		changeSignal(depth->d, dyaddown, LoDFilter_.size()-1);
		// compute left cutoff
		unsigned cutLen = ((int) pow((float) 2, (float) wDepth_ - d - 1) - 1) * (LoDFilter_.size()-1); 
		
		// update for next step
		input = depth->a;
		
		changeSignal(depth->a, leftcut, cutLen);
		changeSignal(depth->d, leftcut, cutLen);
		
		// depending on segment being last
		if(seg.isLast) {
			// compute how many samples will be left untouched
			unsigned toBeLeft = (int) floor((float)(lastSegSize_ + getLeftExtSize(n_)) * (float) pow( (float) 2, (float) -1 * (d+1)));
			toBeLeft -= ((int) pow((float) 2, (float) wDepth_ - d - 1) - 1) * (LoDFilter_.size()-1);
			// so how many to cutOff right?
			unsigned rightCut = (unsigned) (depth->a.size() - toBeLeft);
			changeSignal(depth->a, rightcut, rightCut);
			changeSignal(depth->d, rightcut, rightCut);		
		}
			
		if(DEBUG) {
			std::cout << "Coefs A, depth " << d+1 << ": " << depth->a << std::endl;
			std::cout << "Coefs D, depth " << d+1 << ": " << depth->d << std::endl;
		}
		
		// update for next level

		d++; depth++; 		
	}
}

// convolution
// takes signal and filter to produce result, overwrites result, all params &ref
void SegDWT::conv(vecType& signal, vecType& filter, vecType& result) {
	// init result
	// get sizes
	unsigned fLen = filter.size();
	unsigned xLen = signal.size();
	result.assign(fLen + xLen - 1, 0.0);
	
	// convolution with zero pad
	for(unsigned j=0; j<(fLen + xLen - 1); ++j) {
		for(unsigned i=0; i<fLen; ++i) {
			if(j-i >= 0 && j-i < xLen)
				result[j] += filter[i] * signal[j-i];
		}
	}
}

// changes signal (dyaddown)
// params: &ref to signal, and type of work
// cutOff - if dyaddown, both-side cutoff
//		  - if leftcut, only left-side cutoff
//		  - if rightcut, only right-side cutoff
void SegDWT::changeSignal(vecType& signal, workType type, unsigned cutOff) {
	// init
	unsigned origSize = signal.size();
	vecIter start = signal.begin();
	vecType backup;
	
	// perform
	switch(type) {
		case dyaddown:
			
			// cutoff left
			if(origSize > cutOff && cutOff > 0) {
				start = signal.erase(start, start + cutOff);	
			}
			
			// refresh iterator & size
			// cutoff right
			origSize = signal.size();
			start = signal.begin();
			if(origSize > cutOff-1 && cutOff > 0) {
				signal.erase(start + (origSize-cutOff+1), signal.end());
			}			
			
			// copy-based dyaddown
			backup = signal;
			signal.assign((unsigned) floor((double) signal.size() / (double) 2), 0.0);
			start = signal.begin();
			for(unsigned j=1; j<backup.size(); j+=2) {
				*start++ = backup[j];
				if(start == signal.end())
					break;
			}
			
			break;
			
		case leftcut:
			// cutoff left
			if(origSize > cutOff && cutOff > 0) {
				signal.erase(start, start + cutOff);	
			} else if(cutOff >= origSize) {
				signal.clear();
			}
			break;
			
		case rightcut:
			// cutoff right
			if(origSize > cutOff && cutOff > 0) {
				signal.erase(start + (origSize-cutOff), signal.end());	
			} else if(cutOff >= origSize) {
				signal.clear();
			}
			break;
	}	
}


// one segment constructor
SegDWT::Segment::Segment(unsigned depth, bool lst) : isLast(lst) {
	for(unsigned j=0; j<depth; ++j)
		coefs.push_back(WaveletCoefs());
}

