// definitions for segdwt
#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <sstream>
#include <cmath>
#include "segidwt.h"

// ostream overloads
ostream& operator<<(ostream& stream, vecType& vec) {
	vecIter iter = vec.begin();
	stream << "Vector output (size = " << vec.size() << ")" << std::fixed << std::setprecision(3) << std::endl;
	while(iter != vec.end()) {
		stream << *iter++ << "\t";
	}
	stream << std::endl;	
	return stream; 
}

// ostream overloads
ostream& operator<<(ostream& stream, std::vector<int>& vec) {
	std::vector<int>::iterator iter = vec.begin();
	stream << "Integer vector output (size = " << vec.size() << ")" << std::endl;
	while(iter != vec.end()) {
		stream << *iter++ << "\t";
	}
	stream << std::endl;	
	return stream; 
}




// SegIDWT class public definitions
SegIDWT::SegIDWT(bool debg): debug_(debg) {}

// load filter specified by filename
// exception-based model of error-treating
void SegIDWT::loadFilter(const char *filename) {
	std::fstream file;
	file.open(filename);
	if(!file) {
		std::cout << "Can't open filter file!" << std::endl;
		throw ExcFilterFile();
	}
	
	// process all lines of text
	unsigned filterType = 0;
	std::string line;
	
	while(!std::getline(file, line).eof() && filterType < 4) {
		type value=0.0;
		std::istringstream strCheck(line);
		
		// check text line
		if(strCheck >> value) {
			switch(filterType) {
				case 0:
					LoDFilter_.push_back(value);
					break;
				case 1:
					HiDFilter_.push_back(value);
					break;
				case 2:
					LoRFilter_.push_back(value);
					break;
				case 3:
					HiRFilter_.push_back(value);
					break;
					
			}
		} else {
			filterType++;
		}
	}
	
	// check last vector
	if(HiRFilter_.size() == 0 || LoDFilter_.size() == 0 || LoRFilter_.size() == 0 || HiDFilter_.size() == 0) {
		std::cout << "Filter file in bad format or incomplete!" << std::endl;
		throw ExcFilterFile();
	}
	
	std::cout << "Filter file \"" << filename << "\" loaded. Coefs count: "
				<< "LoD=" << LoDFilter_.size() << ", "
				<< "HiD=" << HiDFilter_.size() << ", "
				<< "LoR=" << LoRFilter_.size() << ", "
				<< "HiR=" << HiRFilter_.size() << std::endl;	  
}

// load signal file specified by filename
// exception-based model of error-treating
void SegIDWT::loadResult(const char *filename) {
	std::fstream file;
	file.open(filename);
	if(!file) {
		std::cout << "Can't open results file!" << std::endl;
		throw ExcSignalFile();
	}
	
	// process all lines of text
	std::string line;
	
	if(!std::getline(file, line).eof()) {
		unsigned value=0, value2=0;
		std::istringstream strCheck(line);
		if(strCheck >> value >> value2) {
			segCount_ = value;
			wDepth_ = value2;
		} else {
			std::cout << "Results file either incomplete or in wrong format!" << std::endl;	
			throw ExcResultFile();	
		}
	} else {
		std::cout << "Results file either incomplete or in wrong format!" << std::endl;
		throw ExcResultFile();
	}
	
	twoToJ_ = (int) pow((float) 2, (float) wDepth_);
	
	// read segment sizes - assume segment size same for all (exc. last!)
	segSize_ = 0; lastSegSize_ = 0;
	unsigned cnt = 0;
	while(!std::getline(file, line).eof()) {
		unsigned value=0;
		std::istringstream strCheck(line);
		// read only segment size
		if(strCheck >> value) {
			if(value != segSize_) {
				if(cnt < segCount_-1) {
					segSize_ = value;
				} else {
					lastSegSize_ = value;
				}
			}
			seg_.push_back(Segment(wDepth_, (lastSegSize_ > 0)?true:false));
		} else {
			// seems like end of bookkeeping matrix
			break;
		}	
		cnt++;
	}
	
	// read segment data
	// init segment iterator
	std::list<Segment>::iterator segmentIt = seg_.begin();
	unsigned s=0;
	while(segmentIt != seg_.end()) {
		// init coef iterator
		std::list<Segment::WaveletCoefs>::iterator coefIt = segmentIt->coefs.begin();
		unsigned d=0;
		while(coefIt != segmentIt->coefs.end()) {
			// compute the zero ADD
			unsigned zeros;
			if(d < wDepth_-1) {
				zeros = ((int) pow((float) 2, (float) wDepth_ - d - 1) - 1) *  (LoDFilter_.size()-1);
			} else {
				zeros = 0;
			}
			
			// for each DETAIL coefs
			if(zeros) {
				coefIt->d.assign(zeros, 0.0);
				std::cout << "Zero-extending D" << d << ", seg. " << s << " coefs by " << zeros << " samples from left." << std::endl;
			}
			while(!std::getline(file, line).eof()) {
				type value=0.0;
				std::istringstream strCheck(line);
				if(strCheck >> value) {
					coefIt->d.push_back(value);	
				} else {
					break;
				}
			}
			d++;
			// for each APPROX coefs
			if(d == wDepth_) {
				while(!std::getline(file, line).eof()) {
					type value=0.0;
					std::istringstream strCheck(line);
					if(strCheck >> value) {
						coefIt->a.push_back(value);	
					} else {
						break;
					}
				}
			}
			coefIt++; 
		}		
		segmentIt++; s++;
	}
	
	file.close();
	std::cout << "File \"" << filename << "\" successfully loaded. Obtained parameters:" << std::endl;
	std::cout << "Depth=" << wDepth_ << ", SegCount=" << segCount_ << ", SegSize=" << segSize_ << ", lastSegSize=" << lastSegSize_ << std::endl  << std::endl; 
	
}
// save the assembled signal
void SegIDWT::saveSignal(const char *filename) {
	std::ofstream file;

	file.open(filename, std::ios::out);
	file << std::fixed << std::setprecision(15);
	vecIter sigIt = signal_.begin();
	while(sigIt != signal_.end()) {
		file << *sigIt << std::endl;
		sigIt++;
	}
	
	file.close();
	std::cout << "Assembled signal saved to file \"" << filename << "\"." << std::endl;
}
// proceed with transform
void SegIDWT::performTransform() {
	// generate extension size, will stay solid
	extSize_ = ((int) pow((float) 2, (float) wDepth_) - 1) * (LoRFilter_.size() - 1);
	int par = 0;
	unsigned s=0;
	
	// storage for additions
	vecType toAdd;
	
	// init segment iterator
	std::list<Segment>::iterator segmentIt = seg_.begin();
	// process all segments	
	while(segmentIt != seg_.end()) {
		// init temp segment space
		vecType seg;
		// perform WT
		mlwr(*segmentIt, seg);
		// get parity
		par = parity(par, (segmentIt->isLast) ? lastSegSize_ : segSize_ );
		
		// branch if last segment
		if(segmentIt->isLast) {
			// generate Pmin vector
 			int vectSize = getPminVectSize(wDepth_, segSize_);
			std::vector<int> pmin;
			genPminValues(pmin, vectSize, segSize_, wDepth_);
			
			if(DEBUG)
				std::cout << "Pmin: " << pmin << std::endl;
			
			// compute index / exactly as MATLAB code
			int index = (s + 1 + 3*vectSize - 2) % vectSize;	
			// estimate difference			
			int diff = pmin[index] - lastSegSize_;
			// max quick replacement
			diff = (diff > par) ? diff : par;
			// estimate length - with both croppings to the right
			// it starts as extension, only if segment is not first
			unsigned endLen = (s==0)? 0 : extSize_;
			// then we take a part of result (first right cut)
			// apply this only if extSize > last seg len
			if(extSize_ < seg.size() - LoRFilter_.size() + 1) {
				endLen += (seg.size() - LoRFilter_.size() + 2 - extSize_);
			}
			// and we apply difference to this value (second right cut)
			endLen -= diff;
			
			// finish the segment 
			if(endLen > 0) {
				// perform toAdd
				unsigned sizeAdd = (extSize_ < endLen) ? extSize_ : endLen;
				if(sizeAdd > toAdd.size())
					sizeAdd = toAdd.size();
				for(unsigned i=0; i<sizeAdd; ++i)
					seg[i] = toAdd[i] + seg[i];
			}
			
			// remove right junk
			//if(endLen > extSize_) {
				changeSignal(seg, rightcut, (seg.size() - endLen));
			//}
		
			// add segment to final signal
			signal_.insert(signal_.end(), seg.begin(), seg.end());
			
			if(DEBUG) {
				std::cout << "Last segment processed. Length of it's result is " << endLen << std::endl;
				std::cout << "Segment [" << s << "] (last) data: " << seg << std::endl;
			}
				
		
		} else {
			// is it possible to extract a centerpart from the signal?
			int centerSize = seg.size() - 2*extSize_;
			vecType centerPart;
			if(centerSize > 0) {
				// store centerPart
				centerPart.assign(seg.begin() + extSize_, seg.end() - extSize_);
			}	
			
			// consider a toAdd
			if(s>0) {
				// just to be bulletproof
				unsigned addSize = (extSize_ > seg.size()) ? seg.size() : extSize_;
				if(addSize > toAdd.size())
					addSize = toAdd.size();
				// now add toAdd to extSize beginning	
				for(unsigned j=0; j<addSize; ++j)
					seg[j] = seg[j] + toAdd[j];
				// keep o	nly extSize + centerSize
				/*if(seg.size() > extSize_ + centerSize) {
					changeSignal(seg, rightcut, (seg.size() - (extSize_ + centerSize)));
				}*/
				// save to signal
				unsigned endIndex = (seg.size() > extSize_ + centerSize) ? extSize_ + centerSize : seg.size();
				signal_.insert(signal_.end(), seg.begin(), seg.begin() + endIndex);
				if(DEBUG)
					std::cout << "Segment [" << s << "] data: " << seg << std::endl;
				
			} else {
				if(centerSize > 0) {
					// add centerpart to the signal
					signal_.insert(signal_.end(), centerPart.begin(), centerPart.end());
					if(DEBUG)
						std::cout << "Segment [" << s << "] data: " << centerPart << std::endl;
				} else {
					if(DEBUG)
						std::cout << "Segment [" << s << "] has no data to join." << std::endl;	
				}
				
			}
			
			// update toAdd buffer with extSize rest of seg
			// first get the size
			unsigned startIdx = ((int) seg.size() - (int) extSize_ > 0) ? ((int) seg.size() - (int) extSize_) : 0;
			toAdd.assign(seg.begin()+startIdx, seg.end());	
			if(DEBUG)
				std::cout << "toAdd: " << toAdd << std::endl;		
		}
		
		// update iterator and segCounter
		s++; segmentIt++;
	}
}


// mlwr algorithm
// params: &ref to Segment, &ref to source &vecType data
void SegIDWT::mlwr(SegIDWT::Segment &seg, vecType &src) {
	// init depth
	unsigned d=wDepth_;
	// reverse iterator
	std::list<Segment::WaveletCoefs>::reverse_iterator depth = seg.coefs.rbegin();
	
	// inital assembly
	vecType assemblyA(depth->a);
	vecType assemblyD;
	
	// for every depth (d=wDepth..bottom, d=0..surface)
	while(d > 0 && depth != seg.coefs.rend()) {
		// get details of this level
		assemblyD = depth->d;
		
		//perform dyadup
		changeSignal(assemblyA, dyadup);
		changeSignal(assemblyD, dyadup);
		
		// create temp vectors for conv result storage
		vecType afterConvA;
		vecType afterConvD;
		
		// convolve		
		conv(assemblyA, LoRFilter_, afterConvA);
		conv(assemblyD, HiRFilter_, afterConvD);
		
		// surface
		if(d == 1) {
			// estimate aprsSize
			unsigned aprsSize = 2 * depth->d.size() + LoRFilter_.size() - 1;
			// perform summation
			src.assign(aprsSize, 0.0);
			for(unsigned j=0; j<aprsSize; ++j)
				src[j] = afterConvA[j] + afterConvD[j];
			// just for the sake of order, update iterator
			depth++;
			
		// or regular depth
		} else {
			// safe to increment, so update iterator
			depth++;
			// we borrow the size from next level details
			unsigned aprsSize = (unsigned) depth->d.size();
			
			// now we perform sumation
			assemblyA.assign(aprsSize, 0.0);
			for(unsigned j=0; j<aprsSize; ++j) {
				type valueA = (j < afterConvA.size())? afterConvA[j] : 0.0;
				type valueD = (j < afterConvD.size())? afterConvD[j] : 0.0;
				assemblyA[j] = valueA + valueD;
			}

			// to be nice, also store the A in the segment data
			depth->a = assemblyA;
		}
		
		// update depthCounter
		d--;
	}
}

// convolution
// takes signal and filter to produce result, overwrites result, all params &ref
void SegIDWT::conv(vecType& signal, vecType& filter, vecType& result) {
	// init result
	// get sizes
	unsigned fLen = filter.size();
	unsigned xLen = signal.size();
	result.assign(fLen + xLen - 1, 0.0);
	
	// convolution with zero pad
	for(unsigned j=0; j<(fLen + xLen - 1); ++j) {
		for(unsigned i=0; i<fLen; ++i) {
			if(j-i >= 0 && j-i < xLen)
				result[j] += filter[i] * signal[j-i];
		}
	}
}

// changes signal (dyaddown)
// params: &ref to signal, and type of work
// cutOff - if dyaddown, both-side cutoff
//		  - if leftcut, only left-side cutoff
//		  - if rightcut, only right-side cutoff
void SegIDWT::changeSignal(vecType& signal, workType type, unsigned cutOff) {
	// init
	unsigned origSize = signal.size();
	vecIter start = signal.begin();
	vecType backup;
	
	// perform
	switch(type) {
		case dyadup:
			// temp buffer dyadup
			backup = signal;
			start = backup.begin();
			signal.assign(origSize * 2, 0.0);
			
			// relatively safe way of interleaving
			for(unsigned j=1; j<signal.size(); j+=2) {
				if(start == backup.end())
					break;
				signal[j] = *start++;
			} 
			
			break;
			
		case leftcut:
			// cutoff left
			if(origSize > cutOff && cutOff > 0) {
				signal.erase(start, start + cutOff);	
			} else if(cutOff >= origSize) {
				signal.clear();
			}
			break;
			
		case rightcut:
			// cutoff right
			if(origSize > cutOff && cutOff > 0) {
				signal.erase(start + (origSize-cutOff), signal.end());	
			} else if(cutOff >= origSize) {
				signal.clear();
			}
			break;
	}	
}

// running parity
int SegIDWT::parity(int last, int number) {
	int newp = (number % 2 > 0);
	return last ^ newp;
}

// greatest common divisor
int SegIDWT::GCD(int a, int b) {
    while( 1 )
    {
        a = a % b;
		if( a == 0 )
			return b;
		b = b % a;
        if( b == 0 )
			return a;
    }
}

// least common multiple
int SegIDWT::LCM(int a, int b) {
	return ( a * b ) / GCD(a,b);
}

// greatest dividing power of two
int SegIDWT::GDPT(int n) {
	int power = 0;
	while(LCM(n,2) == n) {
		power++;
		n /= 2;
	}
	return power;
}

// get Pmin vector size
int SegIDWT::getPminVectSize(int J, int s) {
	int res = GDPT(s);
	if(res > J)
		res = J;

	return (int) pow((float) 2, (float) J - res);
}

// generate Pmin values
void SegIDWT::genPminValues(std::vector<int>& pmin, int size, int s, int J) {
	pmin.assign(size, 0);
	for(int i=0; i<size; ++i)
		pmin[i] = (int) pow((float) 2, (float) J) * (int) ceil((float) ((i+1) * s) / (int) pow((float) 2, (float) J)) - (i+1) * s;
} 


// one segment constructor
SegIDWT::Segment::Segment(unsigned depth, bool lst) : isLast(lst) {
	for(unsigned j=0; j<depth; ++j)
		coefs.push_back(WaveletCoefs());
}

